#include "str8.h"
#include "aliases.h"
#include "mem_allocator.h"
#include <string.h>
#include <stdbool.h>

internal Str8List node_to_list(Str8Node *node);

Str8 *wapp_str8_buf_alloc(const Allocator *allocator, u64 capacity) {
  Str8 *str = NULL;

  if (!allocator) {
    goto RETURN_STR8;
  }

  str = wapp_mem_allocator_alloc(allocator, sizeof(Str8));
  if (!str) {
    goto RETURN_STR8;
  }

  c8 *buf = wapp_mem_allocator_alloc(allocator, sizeof(c8) * capacity);
  if (!buf) {
    wapp_mem_allocator_free(allocator, (void **)&str);
    goto RETURN_STR8;
  }

  str->buf      = buf;
  str->size     = 0;
  str->capacity = capacity;

RETURN_STR8:
  return str;
}

Str8 *wapp_str8_alloc_cstr(const Allocator *allocator, const char *str) {
  Str8 *output = NULL;

  if (!allocator || !str) {
    goto RETURN_ALLOC_CSTR;
  }

  u64 length = strlen(str);
  output     = wapp_str8_buf_alloc(allocator, length * 2);
  if (!output) {
    goto RETURN_ALLOC_CSTR;
  }

  output->size = length;
  memcpy(output->buf, str, length);

RETURN_ALLOC_CSTR:
  return output;
}

Str8 *wapp_str8_alloc_str8(const Allocator *allocator, Str8RO *str) {
  Str8 *output = NULL;

  if (!allocator || !str) {
    goto RETURN_ALLOC_STR8;
  }

  output = wapp_str8_buf_alloc(allocator, str->capacity);
  if (!output) {
    goto RETURN_ALLOC_STR8;
  }

  output->size = str->size;
  memcpy(output->buf, str->buf, str->size);

RETURN_ALLOC_STR8:
  return output;
}

c8 wapp_str8_get(const Str8 *str, u64 index) {
  if (index >= str->size) {
    return '\0';
  }

  return str->buf[index];
}

void wapp_str8_set(Str8 *str, u64 index, c8 c) {
  if (index >= str->size) {
    return;
  }

  str->buf[index] = c;
}

bool wapp_str8_equal(Str8RO *s1, Str8RO *s2) {
  if (s1->size != s2->size) {
    return false;
  }

  return memcmp(s1->buf, s2->buf, s1->size) == 0;
}

Str8RO wapp_str8_substr(Str8RO *str, u64 start, u64 end) {
  if (start >= str->size || start >= end) {
    return (Str8RO){0};
  }

  if (end > str->size) {
    end = str->size;
  }

  return (Str8RO){
    .capacity = end - start,
    .size     = end - start,
    .buf      = str->buf + start,
  };
}

i64 wapp_str8_find(Str8RO *str, Str8RO substr) {
  if (substr.size > str->size) {
    return -1;
  }

  for (u64 i = 0; i < str->size; ++i) {
    const c8 *sub = str->buf + i;
    if (memcmp(sub, substr.buf, substr.size) == 0) {
      return i;
    }
  }

  return -1;
}

i64 wapp_str8_rfind(Str8RO *str, Str8RO substr) {
  if (substr.size > str->size) {
    return -1;
  }

  for (i64 i = str->size - substr.size; i >= 0; --i) {
    const c8 *sub = str->buf + i;
    if (memcmp(sub, substr.buf, substr.size) == 0) {
      return i;
    }
  }

  return -1;
}

Str8Node *wapp_str8_list_get(Str8List *list, u64 index) {
  if (index >= list->node_count) {
    return NULL;
  }

  Str8Node *output  = NULL;
  Str8Node *current = list->first;
  for (u64 i = 1; i <= index; ++i) {
    current = current->next;
  }

  output = current;

  return output;
}

void wapp_str8_list_push_front(Str8List *list, Str8Node *node) {
  if (!list || !node || !(node->string)) {
    return;
  }

  Str8List node_list = node_to_list(node);

  if (list->node_count == 0) {
    *list = node_list;
    return;
  }

  list->total_size += node_list.total_size;
  list->node_count += node_list.node_count;

  Str8Node *first = list->first;
  if (first) {
    first->prev = node_list.last;
  }

  list->first          = node_list.first;
  node_list.last->next = first;
}

void wapp_str8_list_push_back(Str8List *list, Str8Node *node) {
  if (!list || !node || !(node->string)) {
    return;
  }

  Str8List node_list = node_to_list(node);

  if (list->node_count == 0) {
    *list = node_list;
    return;
  }

  list->total_size += node_list.total_size;
  list->node_count += node_list.node_count;

  Str8Node *last = list->last;
  if (last) {
    last->next = node_list.first;
  }

  list->last            = node_list.last;
  node_list.first->prev = last;
}

void wapp_str8_list_insert(Str8List *list, Str8Node *node, u64 index) {
  if (!list || !node || !(node->string)) {
    return;
  }

  if (index == 0) {
    wapp_str8_list_push_front(list, node);
    return;
  } else if (index == list->node_count) {
    wapp_str8_list_push_back(list, node);
    return;
  }

  Str8Node *dst_node = wapp_str8_list_get(list, index);
  if (!dst_node) {
    return;
  }

  Str8List node_list = node_to_list(node);

  list->total_size += node_list.total_size;
  list->node_count += node_list.node_count;

  Str8Node *prev = dst_node->prev;

  dst_node->prev = node_list.last;
  prev->next     = node_list.first;

  node_list.first->prev = prev;
  node_list.last->next  = dst_node;
}

Str8Node *wapp_str8_list_pop_front(Str8List *list) {
  Str8Node *output = NULL;

  if (!list || list->node_count == 0) {
    goto RETURN_STR8_LIST_POP_FRONT;
  }

  output = list->first;

  if (list->node_count == 1) {
    *list = (Str8List){0};
    goto RETURN_STR8_LIST_POP_FRONT;
  }

  --(list->node_count);
  list->total_size -= output->string->size;
  list->first       = output->next;

  output->prev = output->next = NULL;

RETURN_STR8_LIST_POP_FRONT:
  return output;
}

Str8Node *wapp_str8_list_pop_back(Str8List *list) {
  Str8Node *output = NULL;

  if (!list || list->node_count == 0) {
    goto RETURN_STR8_LIST_POP_BACK;
  }

  output = list->last;

  if (list->node_count == 1) {
    *list = (Str8List){0};
    goto RETURN_STR8_LIST_POP_BACK;
  }

  --(list->node_count);
  list->total_size -= output->string->size;
  list->last        = output->prev;

  output->prev = output->next = NULL;

RETURN_STR8_LIST_POP_BACK:
  return output;
}

Str8Node *wapp_str8_list_remove(Str8List *list, u64 index) {
  Str8Node *output = NULL;
  if (!list) {
    goto RETURN_STR8_LIST_REMOVE;
  }

  if (index == 0) {
    output = wapp_str8_list_pop_front(list);
    goto RETURN_STR8_LIST_REMOVE;
  } else if (index == list->node_count) {
    output = wapp_str8_list_pop_back(list);
    goto RETURN_STR8_LIST_REMOVE;
  }

  output = wapp_str8_list_get(list, index);
  if (!output) {
    goto RETURN_STR8_LIST_REMOVE;
  }

  output->prev->next = output->next;
  output->next->prev = output->prev;

  --(list->node_count);
  list->total_size -= output->string->size;

  output->prev = output->next = NULL;

RETURN_STR8_LIST_REMOVE:
  return output;
}

internal Str8List node_to_list(Str8Node *node) {
  Str8List output = {.first = node, .last = node, .total_size = node->string->size, .node_count = 1};

  while (output.first->prev != NULL) {
    output.total_size += output.first->prev->string->size;
    output.first = output.first->prev;
    ++(output.node_count);
  }

  while (output.last->next != NULL) {
    output.total_size += output.last->next->string->size;
    output.last = output.last->next;
    ++(output.node_count);
  }

  return output;
}
